#include "../../ECF/ECF_base.h"
#include "../../ECF/ECF_derived.h"
#include "../../ECF/ECF_macro.h"
#include "AlgBeesAlgorithm.h"

BeesAlgorithm::BeesAlgorithm(){
	name_ = "BeesAlgorithm";
	selWorstOp_ = (SelWorstOpP) (new SelWorstOp);
}

void BeesAlgorithm::registerParameters(StateP state){
	//scouts defines the number of scout bees, default 60
	registerParameter(state, "scouts", (voidP) new uint(60), ECF::INT, "Number of scout bees which the algorithm will use");

	//foragers defines the number of foragers which will be assigned to each best scout, default 40
	registerParameter(state, "foragers", (voidP) new uint(40), ECF::INT, "Number of foragers which  will be assigned to each best scout");

	//local defines the size of the local area search for foragers around the each selected best scout, default 100.0
	registerParameter(state, "local", (voidP) new float(100.0), ECF::FLOAT, "Size of the local area search for foragers around the each selected best scout");
}

bool BeesAlgorithm::initialize(StateP state){
	selWorstOp_->initialize(state);

	voidP par = getParameterValue(state, "scouts");
	scouts_ = *((uint*) par.get());

	par = getParameterValue(state, "foragers");
	foragers_ = *((uint*) par.get());

	par = getParameterValue(state, "local");
	local_ = *((float*) par.get());

	par = state->getRegistry()->getEntry("population.size");
	popSize_ = *((uint*) par.get());

	for(uint i = 0; i < popSize_; i++)
		searchSize_.push_back(local_);

	FloatingPointP flp (new FloatingPoint::FloatingPoint);
	if(state->getGenotypes()[0]->getName() != flp->getName()){
		ECF_LOG_ERROR(state, "Error: Bees algorithm accepts only a single FloatingPoint genotype");
		throw ("");
	}

	dimension_ = *((uint*)((state->getGenotypes()[0]->getParameterValue(state, "dimension")).get()));
	ubound_ = *((double*)((state->getGenotypes()[0]->getParameterValue(state, "ubound")).get()));
	lbound_ = *((double*)((state->getGenotypes()[0]->getParameterValue(state, "lbound")).get()));
	
	return true;
}

bool BeesAlgorithm::advanceGeneration(StateP state, DemeP deme){

	BeesAlgorithm::foragingPhase(state, deme); //searching locally aroung each individual in the current population
	BeesAlgorithm::scoutingPhase(state, deme); //global search and, in case of finding a better solution than the current worst individual in the population,
					// replacing the worst individual in the current population
	return true;
}


void BeesAlgorithm::foragingPhase(StateP state, DemeP deme){
	for(uint i = 0; i < popSize_; i++){
		if(searchSize_[i] < 0.0000000000001)
			continue;

		IndividualP cur = (IndividualP) deme->at(i);

		FloatingPointP flp = boost::static_pointer_cast<FloatingPoint::FloatingPoint> (cur->getGenotype(0));

		IndividualP best;
		best = BeesAlgorithm::createForager(state, cur, searchSize_[0]);
		evaluate(best);
		for(uint j = 1; j < foragers_; j++){
			IndividualP forager = BeesAlgorithm::createForager(state, cur, searchSize_[i]);
			evaluate(forager);
			if(forager->fitness->isBetterThan(best->fitness))
				best = forager;
		}
		if(best->fitness->isBetterThan(cur->fitness)){
			replaceWith(cur, best);
			searchSize_[i] = local_;
		}
		else
			searchSize_[i] *= 1.;
	}
}

IndividualP BeesAlgorithm::createForager(StateP state, IndividualP ref, double searchSize){
	IndividualP forager = (IndividualP) ref->copy();
	FloatingPointP fp = boost::static_pointer_cast<FloatingPoint::FloatingPoint> (forager->getGenotype(0));
	for(uint i = 0; i < dimension_; i++){
		fp->realValue[i] = fp->realValue[i] - searchSize + 2 * searchSize * state->getRandomizer()->getRandomDouble();
	}
	return forager;
}

void BeesAlgorithm::scoutingPhase(StateP state, DemeP deme){
	uint worstIndex = findWorstBee(state, deme);
	for(uint i = 0; i < scouts_; i++){
		IndividualP scout = BeesAlgorithm::createScout(state);
		evaluate(scout);
		if(scout->fitness->isBetterThan(deme->at(worstIndex)->fitness)){
			replaceWith(deme->at(worstIndex), scout);
			worstIndex = findWorstBee(state, deme);
		}
	}
}

uint BeesAlgorithm::findWorstBee(StateP state, DemeP deme){
	uint worstIndex = 0;
	FitnessP fit = deme->at(0)->fitness;
	for(uint i = 1; i < popSize_; i++){
		if(fit->isBetterThan(deme->at(i)->fitness)){
			worstIndex = i;
			fit = deme->at(i)->fitness;
		}
	}
	return worstIndex;
}

IndividualP BeesAlgorithm::createScout(StateP state){
	IndividualP scout = (IndividualP) new Individual(state);
	FloatingPointP fp = boost::static_pointer_cast<FloatingPoint::FloatingPoint> (scout->getGenotype(0));
	for(uint i = 0; i < dimension_; i++)
		fp->realValue[i] = lbound_ + (ubound_ - lbound_) * state->getRandomizer()->getRandomDouble();
	return scout;
}

